package cd.backend.codegen;

import cd.Config;
import cd.Main;
import cd.backend.codegen.RegisterManager.Register;
import cd.ir.Ast;
import cd.ir.Ast.ClassDecl;
import cd.ir.Symbol.*;

import java.io.Writer;
import java.util.*;

import static cd.Config.MAIN;
import static cd.backend.codegen.RegisterManager.BASE_REG;
import static cd.backend.codegen.RegisterManager.STACK_REG;

public class AstCodeGenerator {
	/** Constant representing the boolean TRUE as integer */
	static final int TRUE = 1;
	/** Constant representing the boolean FALSE as integer */
	static final int FALSE = 0;
	/** Size of any variable in assembly
	 * Primitive variables take up 4 bytes (booleans are integers)
	 * Reference variables are a 4 byte pointer
	 */
	static final int VAR_SIZE = 4;

	RegsNeededVisitor rnv;
	
	ExprGenerator eg;
	StmtGenerator sg;
	
	protected final Main main;
	
	final AssemblyEmitter emit;
	final RegisterManager rm = new RegisterManager();

	AstCodeGenerator(Main main, Writer out) {
		initMethodData();
		main.allTypeSymbols = new ArrayList<>();
		
		this.emit = new AssemblyEmitter(out);
		this.main = main;
		this.rnv = new RegsNeededVisitor();

		this.eg = new ExprGenerator(this);
		this.sg = new StmtGenerator(this);
	}

	protected void debug(String format, Object... args) {
		this.main.debug(format, args);
	}

	public static AstCodeGenerator createCodeGenerator(Main main, Writer out) {
		return new AstCodeGenerator(main, out);
	}

	/**
	 * Main method. Causes us to emit x86 assembly corresponding to {@code ast}
	 * into {@code file}. Throws a {@link RuntimeException} should any I/O error
	 * occur.
	 * 
	 * <p>
	 * The generated file will be divided into two sections:
	 * <ol>
	 * <li>Prologue: Generated by {@link #emitPrefix()}. This contains any
	 * introductory declarations and the like.
	 * <li>Body: Generated by {@link ExprGenerator}. This contains the main
	 * method definitions.
	 * </ol>
	 */
	public void go(List<? extends ClassDecl> astRoots) {
		// Find main type and assign to main.mainType
		// Add array types to the types list to lookup later
		for (ClassDecl decl : astRoots) {
			if (decl.name.equals("Main")) {
				main.mainType = decl.sym;
			}
			main.allTypeSymbols.add(new ArrayTypeSymbol(decl.sym));
		}

		main.allTypeSymbols.add(new ArrayTypeSymbol(PrimitiveTypeSymbol.intType));
		main.allTypeSymbols.add(new ArrayTypeSymbol(PrimitiveTypeSymbol.booleanType));
		main.allTypeSymbols.add(new ArrayTypeSymbol(ClassSymbol.objectType));

		emitPrefix();
		for (ClassDecl ast : astRoots) {
			sg.gen(ast);
		}
	}

	private void emitPrefix() {
		// Emit some useful string constants (copied from old HW1 method declaration)
		emit.emitRaw(Config.DATA_STR_SECTION);
		emit.emitLabel("STR_NL");
		emit.emitRaw(Config.DOT_STRING + " \"\\n\"");
		emit.emitLabel("STR_D");
		emit.emitRaw(Config.DOT_STRING + " \"%d\"");

		// Define Object, Object[], int, int[], boolean and boolean[]
		List<TypeSymbol> elementTypes = new ArrayList<>();
		elementTypes.add(ClassSymbol.objectType);
		elementTypes.add(PrimitiveTypeSymbol.intType);
		elementTypes.add(PrimitiveTypeSymbol.booleanType);
		emit.emitRaw(Config.DATA_INT_SECTION);
		for (TypeSymbol type : elementTypes) {
			// type vtable
			emit.emitLabel(Label.type(type));
			emit.emitConstantData("0"); // Supertype (null)
			// array vtable
			emit.emitLabel(Label.type(new ArrayTypeSymbol(type)));
			emit.emitConstantData(Label.type(ClassSymbol.objectType)); // Supertype
			emit.emitConstantData(Label.type(type)); // Element type
		}

		// Emit the new Main().main() code to start the program:

		// 1. Enter TEXT and start the program
		emit.emitRaw(Config.TEXT_SECTION);
		emit.emit(".globl", MAIN);
		emit.emitLabel(MAIN);
		// 1.1. Prepare first frame
		emit.emit("push", BASE_REG);
		emit.emitMove(STACK_REG, BASE_REG);

		// 2. Create main variable
		Ast.NewObject newMain = new Ast.NewObject("Main");
		newMain.type = main.mainType;
		Register mainLocation = eg.visit(newMain, null);

		// 3. Call main()
		emit.emit("push", mainLocation);
		for (ClassSymbol sym = main.mainType; sym != ClassSymbol.objectType; sym = sym.superClass)
			if (sym.methods.getOrDefault("main", null) != null) {
				emit.emit("call", Label.method(sym, sym.methods.get("main")));
				break;
			}
		emitMethodSuffix(true);
	}

	void initMethodData() {
		rm.initRegisters();
	}


	void emitMethodSuffix(boolean returnNull) {
		if (returnNull)
			emit.emit("movl", 0, Register.EAX);
		emit.emitRaw("leave");
		emit.emitRaw("ret");
	}

	static class Label {
		static String type(TypeSymbol symbol) {
			if (symbol instanceof ClassSymbol)
				return String.format("type_%s", symbol);
			else if (symbol instanceof ArrayTypeSymbol)
				return String.format("array_%s", ((ArrayTypeSymbol) symbol).elementType);
			else if (symbol instanceof PrimitiveTypeSymbol)
				return String.format("primive_%s", symbol);
			throw new RuntimeException("Unimplemented type symbol");
		}

		static String method(ClassSymbol classSymbol, MethodSymbol methodSymbol) {
			return String.format("method_%s_%s", classSymbol.name, methodSymbol.name);
		}

		static String returnMethod(ClassSymbol classSymbol, MethodSymbol methodSymbol) {
			return String.format("return_%s", method(classSymbol, methodSymbol));
		}
	}
}
