package cd.backend.codegen;

import cd.backend.ExitCode;
import cd.backend.codegen.AstCodeGenerator.*;
import cd.backend.codegen.RegisterManager.Register;
import cd.ir.Ast.*;
import cd.ir.ExprVisitor;
import cd.ir.Symbol.ClassSymbol;
import cd.ir.Symbol.MethodSymbol;
import cd.ir.Symbol.PrimitiveTypeSymbol;
import cd.ir.Symbol.VariableSymbol;
import cd.util.debug.AstOneLine;

import java.util.*;

import static cd.Config.SCANF;
import static cd.backend.codegen.AstCodeGenerator.*;
import static cd.backend.codegen.RegisterManager.BASE_REG;
import static cd.backend.codegen.RegisterManager.CALLER_SAVE;
import static cd.backend.codegen.RegisterManager.STACK_REG;

/**
 * Generates code to evaluate expressions. After emitting the code, returns a
 * String which indicates the register where the result can be found.
 */
class ExprGenerator extends ExprVisitor<Register,Location> {
	private final AstCodeGenerator cg;

	ExprGenerator(AstCodeGenerator astCodeGenerator) {
		cg = astCodeGenerator;
	}

	public Register gen(Expr ast) {
		return visit(ast, null);
	}

	@Override
	public Register visit(Expr ast, Location arg) {
		try {
			cg.emit.increaseIndent("Emitting " + AstOneLine.toString(ast));
			if (cg.rnv.calc(ast) > cg.rm.availableRegisters()) {
				Deque<Register> pushed = new ArrayDeque<>();
				for (Register r : RegisterManager.GPR) {
					if (cg.rm.isInUse(r)) {
						cg.emit.emit("push", r);
						pushed.push(r);
						cg.rm.releaseRegister(r);
					}
				}
				Register result = super.visit(ast, arg);
				for (Register r : pushed)
					cg.rm.useRegister(r);
				Register finalResult = cg.rm.getRegister();
				cg.emit.emitMove(result, finalResult);
				for (Register r = pushed.pop(); !pushed.isEmpty(); r = pushed.pop())
					cg.emit.emit("pop", r);
				return finalResult;
			} else return super.visit(ast, arg);
		} finally {
			cg.emit.decreaseIndent();
		}

	}

	@Override
	public Register binaryOp(BinaryOp ast, Location arg) {
		// Simplistic HW1 implementation that does
		// not care if it runs out of registers

		int leftRN = cg.rnv.calc(ast.left());
		int rightRN = cg.rnv.calc(ast.right());

		Register leftReg, rightReg;
		if (leftRN > rightRN) {
			leftReg = visit(ast.left(), arg);
			rightReg = visit(ast.right(), arg);
		} else {
			rightReg = visit(ast.right(), arg);
			leftReg = visit(ast.left(), arg);
		}

		cg.debug("Binary Op: %s (%s,%s)", ast, leftReg, rightReg);

		switch (ast.operator) {
		case B_TIMES:
			cg.emit.emit("imul", rightReg, leftReg);
			break;
		case B_PLUS:
			cg.emit.emit("add", rightReg, leftReg);
			break;
		case B_MINUS:
			cg.emit.emit("sub", rightReg, leftReg);
			break;
		case B_DIV:
		case B_MOD:
			// Check division by 0
			String beginDivision = cg.emit.uniqueLabel();
			cg.emit.emit("cmp", 0, rightReg);
			cg.emit.emit("jne", beginDivision);
			Interrupts.exit(cg, ExitCode.DIVISION_BY_ZERO);
			cg.emit.emitLabel(beginDivision);

			// Save EAX, EBX, and EDX to the stack if they are not used
			// in this subtree (but are used elsewhere). We will be
			// changing them.
			List<Register> dontBother = Arrays.asList(rightReg, leftReg);
			Register[] affected = { Register.EAX, Register.EBX, Register.EDX };

			for (Register s : affected)
				if (!dontBother.contains(s) && cg.rm.isInUse(s))
					cg.emit.emit("pushl", s);

			// Move the LHS (numerator) into eax
			// Move the RHS (denominator) into ebx
			cg.emit.emit("pushl", rightReg);
			cg.emit.emit("pushl", leftReg);
			cg.emit.emit("popl", Register.EAX);
			cg.emit.emit("popl", Register.EBX);
			cg.emit.emitRaw("cltd"); // sign-extend %eax into %edx
			cg.emit.emit("idivl", Register.EBX); // division, result into edx:eax

			// Move the result into the LHS, and pop off anything we saved
			if (ast.operator == BinaryOp.BOp.B_DIV)
				cg.emit.emitMove(Register.EAX, leftReg);
			else
				cg.emit.emitMove(Register.EDX, leftReg);
			for (int i = affected.length - 1; i >= 0; i--) {
				Register s = affected[i];
				if (!dontBother.contains(s) && cg.rm.isInUse(s))
					cg.emit.emit("popl", s);
			}
			break;
		case B_AND:
			cg.emit.emit("and", rightReg, leftReg);
			break;
		case B_OR:
			cg.emit.emit("or", rightReg, leftReg);
			break;
		case B_EQUAL: // a == b  <-->  ! (a != b)
			cg.emit.emit("xor", rightReg, leftReg);
			// if 'leftReg'==0, set leftReg to '1'
			String equal = cg.emit.uniqueLabel();
			String end = cg.emit.uniqueLabel();
			cg.emit.emit("cmp", 0, leftReg);
			cg.emit.emit("je", equal);
			cg.emit.emitMove(AssemblyEmitter.constant(0), leftReg);
			cg.emit.emit("jmp", end);
			cg.emit.emitLabel(equal);
			cg.emit.emitMove(TRUE, leftReg);
			cg.emit.emitLabel(end);
			break;
		case B_NOT_EQUAL:
			String skipTrue = cg.emit.uniqueLabel();
			cg.emit.emit("xor", rightReg, leftReg);
			cg.emit.emit("cmp", 0, leftReg);
			cg.emit.emit("je", skipTrue);
			cg.emit.emitMove(TRUE, leftReg);
			cg.emit.emitLabel(skipTrue);
			break;
		default: // Comparison operations
			String endLabel = cg.emit.uniqueLabel();
			cg.emit.emit("cmp", rightReg, leftReg);
			// leftReg - rightReg
			cg.emit.emitMove(TRUE, leftReg);
			switch (ast.operator) {
			case B_LESS_THAN:
				cg.emit.emit("jl", endLabel);
				break;
			case B_LESS_OR_EQUAL:
				cg.emit.emit("jle", endLabel);
				break;
			case B_GREATER_THAN:
				cg.emit.emit("jg", endLabel);
				break;
			case B_GREATER_OR_EQUAL:
				cg.emit.emit("jge", endLabel);
				break;
			default:
				throw new RuntimeException("A binary operation wasn't implemented");
			}
			cg.emit.emitMove(FALSE, leftReg);
			cg.emit.emitLabel(endLabel);
		}

		cg.rm.releaseRegister(rightReg);

		return leftReg;
	}

	@Override
	public Register booleanConst(BooleanConst ast, Location arg) {
		Register reg = cg.rm.getRegister();
		cg.emit.emitMove(ast.value ? TRUE : FALSE, reg);
		return reg;
	}

	@Override
	public Register builtInRead(BuiltInRead ast, Location arg) {
		Register reg = cg.rm.getRegister();
		cg.emit.emit("sub", 16, STACK_REG);
		cg.emit.emit("leal", AssemblyEmitter.registerOffset(8, STACK_REG), reg);
		cg.emit.emitStore(reg, 4, STACK_REG);
		cg.emit.emitStore("$STR_D", 0, STACK_REG);
		cg.emit.emit("call", SCANF);
		cg.emit.emitLoad(8, STACK_REG, reg);
		cg.emit.emit("add", 16, STACK_REG);
		return reg;
	}

	/**
	 * A Cast from one type to another: {@code (typeName)arg}
	 */
	@Override
	public Register cast(Cast ast, Location arg) {
		// 1. Obtain a register with the desired type's address
		Register desiredType = cg.rm.getRegister();
		cg.emit.emit("lea", Label.type(ast.type), desiredType); // lea copies the label's address to the reg

		// 2. Get a reference to the object to be casted, and push a copy for later
		Register runtimeType = visit(ast.arg(), arg);
		cg.emit.emit("push", runtimeType);

		// 3. Go to the type's vtable
		cg.emit.emitLoad(0, runtimeType, runtimeType);
		
		// 4. Runtime type check: recursively go to superType until
		String matchLabel = cg.emit.uniqueLabel();
		String checkSuperTypeLabel = cg.emit.uniqueLabel();
		cg.emit.emitLabel(checkSuperTypeLabel);
		cg.emit.emit("cmpl", desiredType, runtimeType);
		cg.emit.emit("je", matchLabel); // 4.1 It matches: ok
		cg.emit.emitLoad(0, runtimeType, runtimeType); // Go to superclass
		
		cg.emit.emit("cmp", 0, runtimeType); // 4.2 null is reached (super type of Object): error
		cg.emit.emit("jne", checkSuperTypeLabel);
		Interrupts.exit(cg, ExitCode.INVALID_DOWNCAST);
		cg.emit.emitLabel(matchLabel);

		cg.rm.releaseRegister(desiredType);

		// 5. Recover pointer to object and return it
		cg.emit.emit("pop", runtimeType);
		return runtimeType;
	}

	@Override
	public Register index(Index ast, Location arg) {
		boolean obtainReference = arg.isObtainReference();
		String invalidBoundLabel = cg.emit.uniqueLabel();
		String validBoundLabel = cg.emit.uniqueLabel();
		String checkBoundsLabel = cg.emit.uniqueLabel();

		// Obtain pointer to array and index to be accessed
		Register pointer = visit(ast.left(), arg);
		Register index = visit(ast.right(), arg);

		// check for null pointer
		cg.emit.emit("cmp", 0, pointer);
		cg.emit.emit("jne", checkBoundsLabel); // 0 != array
		Interrupts.exit(cg, ExitCode.NULL_POINTER);
		
		// check if index>=0 and index<arraySize
		Register arraySizeReg = cg.rm.getRegister();
		cg.emit.emitLabel(checkBoundsLabel);
		cg.emit.emitLoad(1 * VAR_SIZE, pointer, arraySizeReg);
		cg.emit.emit("cmp", 0, index);
		cg.emit.emit("jl", invalidBoundLabel); // 0 > index
		cg.emit.emit("cmp", index, arraySizeReg);
		cg.emit.emit("jg", validBoundLabel); // index < array.length
		cg.emit.emitLabel(invalidBoundLabel);
		Interrupts.exit(cg, ExitCode.INVALID_ARRAY_BOUNDS);
		cg.rm.releaseRegister(arraySizeReg);

		// return array element (base + 2 * VAR_SIZE + index * VAR_SIZE)
		cg.emit.emitLabel(validBoundLabel);
		cg.emit.emit("imul", VAR_SIZE, index);
		cg.emit.emit("add", pointer, index);
		cg.rm.releaseRegister(pointer);
		if (!obtainReference)
			cg.emit.emitLoad(2 * VAR_SIZE, index, index);
		else
			cg.emit.emit("add", 2 * VAR_SIZE, index);
		return index;
	}

	@Override
	public Register intConst(IntConst ast, Location arg) {
		Register reg = cg.rm.getRegister();
		cg.emit.emitMove(ast.value, reg);
		return reg;
	}

	@Override
	public Register field(Field ast, Location arg) {
		boolean obtainReference = arg.isObtainReference();
		String accessFieldLabel = cg.emit.uniqueLabel();
		Register pointer = visit(ast.arg(), arg);
		cg.emit.emit("cmp", 0, pointer);
		cg.emit.emit("jne", accessFieldLabel);
		Interrupts.exit(cg, ExitCode.NULL_POINTER);

		cg.emit.emitLabel(accessFieldLabel);
		String foundFieldLabel = cg.emit.uniqueLabel();
		String lookForFieldLabel = cg.emit.uniqueLabel();
		cg.emit.emit("add", VAR_SIZE, pointer);
		cg.emit.emitLabel(lookForFieldLabel);
		cg.emit.emit("cmpl", ast.sym.hashCode(), AssemblyEmitter.registerOffset(0, pointer));
		cg.emit.emit("je", foundFieldLabel);
		cg.emit.emit("add", 2 * VAR_SIZE, pointer);
		cg.emit.emit("jmp", lookForFieldLabel);
		cg.emit.emitLabel(foundFieldLabel);
		if (obtainReference)
			cg.emit.emit("add", VAR_SIZE, pointer);
		else
			cg.emit.emitLoad(VAR_SIZE, pointer, pointer);
		return pointer;
	}

	/**
	 * Array structure:
	 * <ul>
	 *     <li>Type of array (for casts, assignments and equalities)</li>
	 *     <li>Size of array (N)</li>
	 *     <li>element 0 or pointer to element 0</li>
	 *     <li>...</li>
	 *     <li>element N or pointer to element N</li>
	 * </ul>
	 * The type of the elements is stored in the vtable for the array (in order to 
	 * save space). <br/>
	 *
	 * The pointer that references the array points to the type of the array.
	 * The reason for this is so that all reference types have at pointer + 0
	 * the type, for dynamic type comparisons for assignments, casts and ==/!= operators
	 */
	@Override    
	public Register newArray(NewArray ast, Location arg) {
		String validArraySizeLabel = cg.emit.uniqueLabel();
		
		// Check size of array is positive
		Register arraySizeReg = visit(ast.arg(), arg);
		cg.emit.emit("cmp", 0, arraySizeReg);
		cg.emit.emit("jns", validArraySizeLabel); // size >= 0
		Interrupts.exit(cg, ExitCode.INVALID_ARRAY_SIZE);

		// Reserve for length + 2 variables
		cg.emit.emitLabel(validArraySizeLabel);
		cg.emit.emit("push", arraySizeReg);
		cg.emit.emit("add", 2, arraySizeReg);
		Register arrayPointerReg = calloc(arraySizeReg);
		
		// Store overhead information
		// Type reference
		Register arrayTypeReg = cg.rm.getRegister();
		cg.emit.emit("lea", Label.type(ast.type), arrayTypeReg);
		cg.emit.emitStore(arrayTypeReg, 0 * VAR_SIZE, arrayPointerReg);
		cg.rm.releaseRegister(arrayTypeReg);
		// Number of elements
		Register numElemReg = cg.rm.getRegister();
		cg.emit.emit("pop", numElemReg);
		cg.emit.emitStore(numElemReg, 1 * VAR_SIZE, arrayPointerReg);
		cg.rm.releaseRegister(numElemReg);

		return arrayPointerReg;
	}
	
	/**
	 * Structure of reference type objects (except arrays)
	 * <ul>
	 * 	<li>Pointer to type vtable</li>
	 * 	<li>hashCode0, field0</li>
	 * 	<li>...</li>
	 * 	<li>hashCodeN, fieldN</li>
	 * </ul>
	 * The object pointer points to the first element, and every element
	 * is of VAR_SIZE
	 */
	@Override
	public Register newObject(NewObject ast, Location arg) {
		ClassSymbol sym = (ClassSymbol) ast.type; 

		// Obtain from type size of allocated object and allocate memory
		int size = 1;
		ClassSymbol auxSym = sym;
		while (auxSym != ClassSymbol.objectType) {
			size += auxSym.fields.size() * 2;
			auxSym = auxSym.superClass;
		}
		Register sizeReg = cg.rm.getRegister();
		cg.emit.emitMove(size, sizeReg);

		Register pointer = calloc(sizeReg);
		
		// Store the pointer to the type vtable
		Register auxReg = cg.rm.getRegister();
		cg.emit.emit("lea", Label.type(sym), auxReg);
		cg.emit.emitStore(auxReg, 0, pointer);
		cg.emit.emit("push", pointer); // Save the pointer to the beginning to return later
		cg.rm.releaseRegister(auxReg);

		// Store the hashes for the fields' variable symbols
		cg.emit.emit("add", VAR_SIZE, pointer);
		auxSym = sym;
		while (auxSym != ClassSymbol.objectType) {
			for (VariableSymbol field : auxSym.fields.values()) {
				cg.emit.emitStore(field.hashCode(), 0, pointer);
				cg.emit.emit("add", 2 * VAR_SIZE, pointer);
			}
			auxSym = auxSym.superClass;
		}

		// Recover the initial address and return
		cg.emit.emit("pop", pointer);
		return pointer;
	}

	@Override
	public Register nullConst(NullConst ast, Location arg) {
		Register reg = cg.rm.getRegister();
		cg.emit.emitMove(0, reg);
		return reg;
	}

	@Override
	
	public Register thisRef(ThisRef ast, Location arg) {
		Register register = cg.rm.getRegister();
		cg.emit.emitLoad(2 * VAR_SIZE, BASE_REG, register);
		return register;
	}


	/**
	 * implementation of x86 calling convention according to:
	 * http://unixwiz.net/techtips/win32-callconv-asm.html
	 * following the __cdecl calling convention
	 */
	@Override
	public Register methodCall(MethodCallExpr ast, Location arg) {
		// 0. Save registers to stack
		List<Register> callerSaved = new ArrayList<>();
		for (Register reg : CALLER_SAVE) {
			if (cg.rm.isInUse(reg)) {
				callerSaved.add(0, reg);
				cg.emit.emit("push", Register.EAX);
			}
		}

		// 1. Evaluate all the arguments left-to-right (according to Java's spec)
		// and push them to the stack in right-to-left order. We will push them ltr
		// and then swap them so that the order is reversed
		for (Expr allArgument : ast.allArguments()) {
			Register argumentRegister = visit(allArgument, arg);
			cg.emit.emit("push", argumentRegister);
			cg.rm.releaseRegister(argumentRegister);
		}
		for (int i = 0; i < ast.allArguments().size() / 2; i++) {
			int offset1 = i * VAR_SIZE;
			int offset2 = (ast.allArguments().size() - 1 - i) * VAR_SIZE;
			Register aux1 = cg.rm.getRegister();
			Register aux2 = cg.rm.getRegister();
			cg.emit.emitLoad(offset1, STACK_REG, aux1);
			cg.emit.emitLoad(offset2, STACK_REG, aux2);
			cg.emit.emitStore(aux1, offset2, STACK_REG);
			cg.emit.emitStore(aux2, offset1, STACK_REG);
			cg.rm.releaseRegister(aux1);
			cg.rm.releaseRegister(aux2);
		}

		// 2. Call function
		// 2.1. Search for the method pointer using the hashCode
		Register thisRef = cg.rm.getRegister();
		cg.emit.emitLoad(0, STACK_REG, thisRef);
		int hashCode = ast.methodName.hashCode();
		String lookForMethod = cg.emit.uniqueLabel();
		String methodFound = cg.emit.uniqueLabel();
		cg.emit.emitLoad(0, thisRef, thisRef); // Go to type vtable
		cg.emit.emit("add", 1 * VAR_SIZE, thisRef); // Skip the reference to superType
		cg.emit.emitLabel(lookForMethod);
		cg.emit.emit("cmpl", hashCode, AssemblyEmitter.registerOffset(0, thisRef));
		cg.emit.emit("je", methodFound); // hashCode == methodName.hashCode
		cg.emit.emit("add", 2 * VAR_SIZE, thisRef); // Go to next hashCode
		cg.emit.emit("jmp", lookForMethod);
		cg.emit.emitLabel(methodFound);
		// 2.2. Call the function
		cg.emit.emit("call", AssemblyEmitter.registerOffset(VAR_SIZE, thisRef));
		cg.rm.releaseRegister(thisRef);

		// 3. Pop arguments from stack
		cg.emit.emit("add", VAR_SIZE * ast.allArguments().size(), STACK_REG);

		// 4. Return result to caller
		Register result = null;
		if (ast.sym.returnType != PrimitiveTypeSymbol.voidType) {
			result = cg.rm.getRegister();
			cg.emit.emitMove(Register.EAX, result);
		}

		// 5. Restore registers
		for (Register reg : callerSaved) {
			if (cg.rm.isInUse(reg))
				cg.emit.emit("pop", Register.EAX);
			else
				cg.emit.emit("add", 4, Register.ESP); // Don't overwrite
		}
		return result;
	}

	@Override
	public Register unaryOp(UnaryOp ast, Location arg) {
		Register argReg = visit(ast.arg(), arg);
		switch (ast.operator) {
		case U_PLUS:
			break;

		case U_MINUS:
			cg.emit.emit("negl", argReg);
			break;

		case U_BOOL_NOT:
			cg.emit.emit("negl", argReg);
			cg.emit.emit("incl", argReg);
			break;
		}
		return argReg;
	}

	/**
	 * Obtains the value or reference to a variable described by a Var node
	 * and a VariableSymbol.
	 */
	@Override
	public Register var(Var ast, Location arg) {
		boolean obtainReference = arg.isObtainReference();
		Register register = cg.rm.getRegister();
		cg.emit.emitMove(BASE_REG, register);
		int offset;
		switch (ast.sym.kind) {
			case PARAM:
				offset = 3 + arg.methodSym().parameters.indexOf(ast.sym);
				break;
			case LOCAL:
				offset = -(1 + positionOfLocal(arg.methodSym(), ast.name));
				break;
			case FIELD:
				String foundFieldLabel = cg.emit.uniqueLabel();
				String lookForFieldLabel = cg.emit.uniqueLabel();
				cg.emit.emitLoad(2 * VAR_SIZE, register, register);
				cg.emit.emit("add", VAR_SIZE, register);
				cg.emit.emitLabel(lookForFieldLabel);
				cg.emit.emit("cmpl", ast.sym.hashCode(), AssemblyEmitter.registerOffset(0, register));
				cg.emit.emit("je", foundFieldLabel);
				cg.emit.emit("add", 2 * VAR_SIZE, register);
				cg.emit.emit("jmp", lookForFieldLabel);
				cg.emit.emitLabel(foundFieldLabel);
				offset = 1; // The next element will be the reference we want
				break;
			default:
				throw new RuntimeException("VariableSymbol Kind option not implemented");
		}
		if (obtainReference)
			cg.emit.emit("add", offset * VAR_SIZE, register);
		else
			cg.emit.emitLoad(offset * VAR_SIZE, register, register);
		return register;
	}

	private int positionOfLocal(MethodSymbol sym, String name) {
		List<String> locals = new ArrayList<>(sym.locals.keySet());
		locals.sort(Comparator.comparing(o -> o));
		for (int i = 0; i < locals.size(); i++)
			if (locals.get(i).equals(name))
				return i;
		throw new RuntimeException("The variable could not be found, and should be there");
	}

	/**
	 * Helper function that calls calloc to obtain memory for reference type variables
	 * @param numberOfElements Size in bytes, stored in a register
	 * @return A register with the first address of the memory allocated
	 */
	private Register calloc(Register numberOfElements) {
		cg.emit.emitComment("calloc begin arg=" + numberOfElements);
		String callocOk = cg.emit.uniqueLabel();
				
		boolean eaxInUse = cg.rm.isInUse(Register.EAX);
		if (eaxInUse && Register.EAX != numberOfElements)
			cg.emit.emit("push", Register.EAX);		

		// push arguments to the stack, then call calloc
		// the size of allocated elements is always four bytes!
		cg.emit.emit("push", AssemblyEmitter.constant(VAR_SIZE));
		cg.emit.emit("push", numberOfElements);
		cg.rm.releaseRegister(numberOfElements);
		cg.emit.emit("call", "calloc");
		cg.emit.emit("add", 8, STACK_REG);

		// Check for null pointer (if calloc fails)
		cg.emit.emit("cmp", 0, Register.EAX);
		cg.emit.emit("jne", callocOk);
		Interrupts.exit(cg, ExitCode.INTERNAL_ERROR);

        cg.emit.emitLabel(callocOk);
        Register result = cg.rm.getRegister();
        cg.emit.emitMove(Register.EAX, result);

		// pop EAX if needed
		if (eaxInUse && Register.EAX != numberOfElements)
			cg.emit.emit("pop", Register.EAX);

		cg.emit.emitComment("calloc end");
		return result;
	}
}
