package cd.backend.codegen;

import cd.Config;
import cd.backend.ExitCode;
import cd.backend.codegen.AstCodeGenerator.Label;
import cd.backend.codegen.RegisterManager.Register;
import cd.frontend.semantic.ReturnCheckerVisitor;
import cd.ir.Ast;
import cd.ir.Ast.*;
import cd.ir.AstVisitor;
import cd.ir.Symbol;
import cd.ir.Symbol.ClassSymbol;
import cd.ir.Symbol.MethodSymbol;
import cd.util.debug.AstOneLine;

import java.util.*;

import static cd.backend.codegen.AstCodeGenerator.VAR_SIZE;
import static cd.backend.codegen.AstCodeGenerator.TRUE;
import static cd.backend.codegen.RegisterManager.BASE_REG;
import static cd.backend.codegen.RegisterManager.STACK_REG;

/**
 * Generates code to process statements and declarations.
 */
class StmtGenerator extends AstVisitor<Register,Location> {
	protected final AstCodeGenerator cg;

	StmtGenerator(AstCodeGenerator astCodeGenerator) {
		cg = astCodeGenerator;
		
	}

	public void gen(Ast ast) {
		visit(ast, null);
	}

	@Override
	public Register visit(Ast ast, Location arg) {
		try {
			cg.emit.increaseIndent("Emitting " + AstOneLine.toString(ast));
			if (cg.rnv.calc(ast) > cg.rm.availableRegisters()) {
				Deque<Register> pushed = new ArrayDeque<>();
				for (Register r : RegisterManager.GPR) {
					if (cg.rm.isInUse(r)) {
						cg.emit.emit("push", r);
						pushed.push(r);
						cg.rm.releaseRegister(r);
					}
				}
				Register result = super.visit(ast, arg);
				for (Register r : pushed)
					cg.rm.useRegister(r);
				Register finalResult = cg.rm.getRegister();
				cg.emit.emitMove(result, finalResult);
				for (Register r = pushed.pop(); !pushed.isEmpty(); r = pushed.pop())
					cg.emit.emit("pop", r);
				return finalResult;
			} else return super.visit(ast, arg);
		} finally {
			cg.emit.decreaseIndent();
		}
	}

	@Override
	public Register methodCall(MethodCall ast, Location arg) {
		Register result = cg.eg.visit(ast.getMethodCallExpr(), arg);
		if (result != null)
			cg.rm.releaseRegister(result);
		return null;
	}

	public Register methodCall(MethodSymbol sym, List<Expr> allArguments) {
		throw new RuntimeException("Not required");
	}

	/**
	 * vtable structure for a class type: pointer to superclass, and methods, each with 2 entries, name's hashCode and
	 * pointer to method execution
	 */
	@Override
	public Register classDecl(ClassDecl ast, Location arg) {
		cg.emit.emitRaw(Config.DATA_INT_SECTION);
		// Emit vtable for class
		cg.emit.emitLabel(Label.type(ast.sym)); // Label
		cg.emit.emitConstantData(Label.type(ast.sym.superClass)); // Superclass
		// Methods (don't write those that are overridden twice)
		Set<String> generated = new HashSet<>();
		ClassSymbol currSym = ast.sym;
		while (currSym != ClassSymbol.objectType) {
			ClassSymbol finalCurrSym = currSym;
			currSym.methods.values().forEach(o -> {
				if (!generated.add(o.name)) return;
				cg.emit.emitConstantData(String.valueOf(o.name.hashCode()));
				cg.emit.emitConstantData(Label.method(finalCurrSym, o));
			});
			currSym = currSym.superClass;
		}
		// End of class vtable
		// Array vtable
		boolean found = false;
		for (Symbol.TypeSymbol type : cg.main.allTypeSymbols) {
			if (type instanceof Symbol.ArrayTypeSymbol) {
				Symbol.ArrayTypeSymbol aType = (Symbol.ArrayTypeSymbol) type;
				if (aType.elementType == ast.sym) {
					cg.emit.emitLabel(Label.type(aType)); // Label
					found = true;
					break;
				}
			}
		}
		if (!found)
			throw new RuntimeException("The array type could not be found");
		cg.emit.emitConstantData(Label.type(ClassSymbol.objectType)); // Supertype
		cg.emit.emitConstantData(Label.type(ast.sym)); // Type of elements
		// End of array vtable
		cg.emit.emitRaw(Config.TEXT_SECTION);
		
		// Method bodies
		for (Ast method : ast.methods())
			visit(method, new Location(ast.sym));
		
		return null;
	}

	@Override
	public Register methodDecl(MethodDecl ast, Location arg) {
		// Bookkeeping for framework
		arg.enterMethod(ast.sym);
		cg.initMethodData();

		// Begin method
		cg.emit.emitLabel(Label.method(arg.classSym(), ast.sym));

		// 1. Save and update the base register, reserve space for local variables
		cg.emit.emit("enter", ast.sym.locals.size() * VAR_SIZE, 0);
		for (int i = 0; i < ast.sym.locals.size(); i++)
			cg.emit.emitStore(0, -(i + 1) * VAR_SIZE, BASE_REG);

		// 2. Save CPU registers
		Register[] regs = RegisterManager.CALLEE_SAVE;
		for (int i = 0; i < regs.length; i++)
			cg.emit.emit("push", regs[i]);

		// 3. Run the code contained in the function
		Register returnReg = visit(ast.body(), arg);
		cg.emit.emitLabel(Label.returnMethod(arg.classSym(), ast.sym));

		// 4. Restore saved registers (if there is a mismatch, the stack will be corrupted)
		for (int i = regs.length - 1; i >= 0; i--)
			cg.emit.emit("pop", regs[i]);

		// 5 and 6: Restore the old base pointer and return from the function
		cg.emitMethodSuffix(returnReg == null);

		// Framework bookkeeping
		arg.leaveMethod();
		return null;
	}

	@Override
	public Register ifElse(IfElse ast, Location arg) {
		// Emit condition check
		Register conditionRegister = cg.eg.visit(ast.condition(), arg);
		// If both blocks are empty, no more code need be generated, and the condition can be ignored
		if (ast.then().children().isEmpty() && ast.otherwise().children().isEmpty()) {
			// Generate the condition and ignore the result
			cg.rm.releaseRegister(conditionRegister);
			return null;
		}
		String notIfLabel = cg.emit.uniqueLabel();
		String endLabel = cg.emit.uniqueLabel();

		cg.emit.emit("cmp", TRUE, conditionRegister);
		cg.rm.releaseRegister(conditionRegister);
		if (!ast.then().children().isEmpty()) {
			cg.emit.emit("jne", notIfLabel);
			visit(ast.then(), arg);
			// If there is no otherwise, the jump instruction makes no sense
			// as the next code executed will be the next instruction
			if (!ast.otherwise().children().isEmpty())
				cg.emit.emit("jmp", endLabel);
		} else {
			// No if, therefore the else follows the condition immediately
			cg.emit.emit("je", endLabel);
		}

		cg.emit.emitLabel(notIfLabel);
		// Emit otherwise
		visit(ast.otherwise(), arg);
		cg.emit.emitLabel(endLabel);

		// Check if the ifElse ast node contains a return statement
		ReturnCheckerVisitor rc = new ReturnCheckerVisitor();
		if (rc.ifElse(ast, null)) {
			return Register.EAX;
		} else {
			return null;
		}
	}
	
	@Override
	public Register whileLoop(WhileLoop ast, Location arg) {
		String conditionLabel = cg.emit.uniqueLabel();

		cg.emit.emitLabel(conditionLabel);
		Register conditionRegister = cg.eg.visit(ast.condition(), arg);
		cg.emit.emit("cmp", TRUE, conditionRegister);
		cg.rm.releaseRegister(conditionRegister);
		if (ast.body().children().isEmpty()) {
			// Jump structure can be easier if there is no body
			cg.emit.emit("je", conditionLabel);
		} else {
			String endLabel = cg.emit.uniqueLabel();
			// Emit jumps, labels and body
			cg.emit.emit("jne", endLabel);
			visit(ast.body(), arg);
			cg.emit.emit("jmp", conditionLabel);
			cg.emit.emitLabel(endLabel);
		}
		return null;
	}
	
	@Override
	public Register assign(Assign ast, Location arg) {
		Register value = cg.eg.visit(ast.right(), arg);

		// If the type is a reference, visiting the lhs will yield its address
		// else, the lhs will yield the current value. We need a visitor that
		// returns a pointer to the position of the variable/field in memory
		// for primitive types.
		Register pointer = cg.eg.visit(ast.left(), arg.obtainReference());
		if (ast.left().type.isReferenceType()) {
			// Check null pointer
			String validPointerLabel = cg.emit.uniqueLabel();
			cg.emit.emit("cmp", 0, pointer);
			cg.emit.emit("jne", validPointerLabel);
			Interrupts.exit(cg, ExitCode.NULL_POINTER);
			cg.emit.emitLabel(validPointerLabel);
		}
		cg.emit.emitStore(value, 0, pointer);
		cg.rm.releaseRegister(pointer);
		cg.rm.releaseRegister(value);
		return null;
	}

	@Override
	public Register builtInWrite(BuiltInWrite ast, Location arg) {
		Register reg = cg.eg.visit(ast.arg(), arg);
		cg.emit.emit("sub", 16, STACK_REG);
		cg.emit.emitStore(reg, 4, STACK_REG);
		cg.emit.emitStore("$STR_D", 0, STACK_REG);
		cg.emit.emit("call", Config.PRINTF);
		cg.emit.emit("add", 16, STACK_REG);
		cg.rm.releaseRegister(reg);
		return null;
	}

	@Override
	public Register builtInWriteln(BuiltInWriteln ast, Location arg) {
		cg.emit.emit("sub", 16, STACK_REG);
		cg.emit.emitStore("$STR_NL", 0, STACK_REG);
		cg.emit.emit("call", Config.PRINTF);
		cg.emit.emit("add", 16, STACK_REG);
		return null;
	}

	@Override
	public Register returnStmt(ReturnStmt ast, Location arg) {
		if (ast.arg() != null) {
			Register retReg = cg.eg.visit(ast.arg(), arg);
			cg.emit.emitMove(retReg, Register.EAX);
			cg.rm.releaseRegister(retReg);
		}
		cg.emit.emit("jmp", Label.returnMethod(arg.classSym(), arg.methodSym()));
		return Register.EAX;
	}

	@Override
	public Register seq(Seq ast, Location arg) {
		for (Ast instruction : ast.children()) {
			Register res = visit(instruction, arg);
			if (res != null)
				return res; // don't generate instructions after a return
		}
		return null;
	}
}
